/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "RemoteDPPHistogram.h"
#include <common/base/ServerProperties.h>
#include <boost/thread.hpp>
#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/property_tree/json_parser.hpp>
#include <cstdlib>

namespace npp {

using namespace std;
using namespace common;
using namespace boost;
using namespace cameo;
using namespace lstdpp;

using boost::property_tree::ptree;

const std::string RemoteDPPHistogram::TYPE = "remote_dpp_histogram";

RemoteDPPHistogram::RemoteDPPHistogram(const std::string& name) :
	RemoteDPPLiveProcess(name) {

	serverEndpoint.init(this, SAVE, "cameo_server");
	triggerCrate.init(this, SAVE, "trigger_crate");
	triggerBoard.init(this, SAVE, "trigger_board");
	triggerChannel.init(this, SAVE, "trigger_channel");
	delay.init(this, SAVE, "delay");
	dT.init(this, SAVE, "dT");
	maxBlock.init(this, SAVE, "max_block");
	blockSize.init(this, SAVE, "block_size");
	test.init(this, SAVE, "test");
	testFileName.init(this, SAVE, "test_file_name");
	refreshTimeS.init(this, SAVE, "refresh_time_s");
	totalNumberOfChannels.init(this, NOSAVE, "total_number_of_channels");

	crateId.init(this, NOSAVE, "crate_id");
	boardId.init(this, NOSAVE, "board_id");
	channelId.init(this, NOSAVE, "channel_id");
	channelRate.init(this, NOSAVE, "channel_rate");

	xData.init(this, NOSAVE, "x_data");
	histogram.init(this, NOSAVE, "histogram");
	histogramName.init(this, NOSAVE, "histogram_name");
	xLegend.init(this, NOSAVE, "x_legend");
	yLegend.init(this, NOSAVE, "y_legend");
}

RemoteDPPHistogram::RemoteDPPHistogram(const RemoteDPPHistogram& controller) :
	RemoteDPPLiveProcess(controller) {
}

RemoteDPPHistogram::~RemoteDPPHistogram() {
}

void RemoteDPPHistogram::postConfiguration() {

	if (serverEndpoint() == "") {
		m_server.reset(new Server(application::This::getServer().getEndpoint()));

	} else {
		m_server.reset(new Server(serverEndpoint()));
	}

	resetData();
}

void RemoteDPPHistogram::setDetectorControllers(const std::vector<ControllerPtr<acquisition::DetectorController> > & detectorControllers) {

	DPPLiveProcess::setDetectorControllers(detectorControllers);

	// Register the list mode data
	int32 index = 0;
	for (vector<ControllerPtr<acquisition::DetectorController> >::const_iterator c = detectorControllers.begin(); c != detectorControllers.end(); ++c) {
		// Force the attachment to receive updates
		(*c)->attach(this);
		registerUpdater((*c)->listModeData, &RemoteDPPHistogram::updateListModeData, this, index);
		index++;
	}
}

void RemoteDPPHistogram::start() {

	auto_ptr<lstdpp::Reader> reader = initTest();

	resetData();

	// Numor
	string numor;
	size_t size = fileName().size();

	if (size < 6) {
		string prefix(6 - size, '0');
		numor = prefix + fileName();
	} else {
		numor = fileName();
	}

	// Setting the parameters
	ptree parameters;

	parameters.put("numor", numor);
	parameters.put("crate", triggerCrate());
	parameters.put("board", triggerBoard());
	parameters.put("channel", triggerChannel());
	parameters.put("delay", delay() / 10);
	parameters.put("dT", dT() / 10);
	parameters.put("maxBlocks", maxBlock());
	parameters.put("blockSize", blockSize());
	parameters.put("refreshTime", refreshTimeS());

	ptree context = writeListModeContextToPropertyTree(listModeContext);

	parameters.put_child("context", context);

	// Serialize the parameters
	stringstream os;
	write_json(os, parameters, false);


	// Kill the application if one remains
	m_server->killAllAndWaitFor("nhisto");

	vector<string> args;
	args.push_back(os.str());

	// Start the application
	m_histoApplication = m_server->start("nhisto", args);

	// Test if application started
	if (!m_histoApplication->exists()) {
		log(Level::s_Error) << "cannot start histogram application" << endlog;
		return;
	}

	// Create the publisher after the start of the application, because it is a blocking call
	m_publisher.reset();
	m_publisher = application::Publisher::create("dpp_publisher", 1);

	cout << "publisher " << *m_publisher << endl;

	// Wait for the subscriber
	bool sync = m_publisher->waitForSubscribers();

	// Create the subscriber after the publisher
	auto_ptr<application::Subscriber> subscriber = application::Subscriber::create(*m_histoApplication, "dpp_histogram_results");
	cout << "subscriber " << *subscriber << endl;

	// Start the test thread if necessary
	auto_ptr<boost::thread> thread = startTest(reader.get());

	// Listen to the result stream
	vector<double> channelRates;
	while (subscriber->receive(channelRates)) {

		// Copying the rates
		for (int32 i = 0; i < totalNumberOfChannels(); i++) {
			channelRate.set(i, channelRates[i]);
		}

		vector<int32_t> histograms;
		if (!subscriber->receive(histograms)) {
			break;
		}

		// Copying the data
		vector<int32_t>::const_iterator h = histograms.begin();
		for (int32 i = 0; i < totalNumberOfChannels(); i++) {

			if (h == histograms.end()) {
				cerr << "problem with the histogram data" << endl;
				break;
			}

			int32 size = histogram.getSize(i);
			int32 * data = histogram(i);

			for (int32 j = 0; j < size; ++j) {
				if (h == histograms.end()) {
					cerr << "problem with the histogram data" << endl;
					break;
				}

				data[j] = *h;
				++h;
			}

			if (h == histograms.end()) {
				// ok
			}

			histogram.sendEvent(i);
		}
	}

	// Wait for the termination of the application
	application::State state = m_histoApplication->waitFor();

	cout << "histogram application terminated with state " << state << endl;

	// Join the test thread
	finishTest(thread);
}

void RemoteDPPHistogram::stop() {

	// Test the histogram application and the publisher
	if (m_histoApplication.get() != 0 && m_publisher.get() != 0) {
		cout << *m_publisher << " end of stream" << endl;
		m_publisher->sendEnd();
		// We do not stop the application because the termination of the publisher will do it.
	}
}

void RemoteDPPHistogram::raz() {
	resetData();
}

void RemoteDPPHistogram::updateListModeData(int index) {

	if (m_publisher.get() != 0 && !m_publisher->hasEnded()) {
		int32 size = detectorControllers[index]->listModeData.getSize();
		int32 * memoryBlock = detectorControllers[index]->listModeData();

		// Put the array into a string
		string data(reinterpret_cast<char *>(memoryBlock), sizeof(int32) * size);

		// Send the data
		m_publisher->send(data);
	}
}

auto_ptr<lstdpp::Reader> RemoteDPPHistogram::initTest() {

	if (test()) {
		auto_ptr<lstdpp::Reader> reader(new lstdpp::Reader(blockSize()));

		reader->open(testFileName());

		// Override real values
		numberOfADCs = 0;
		numberOfChannels = 0;

		return reader;
	}

	return auto_ptr<lstdpp::Reader>();
}

void RemoteDPPHistogram::testLoop(lstdpp::Reader * reader) {

	if (test()) {
		// Loop over events
		while (!m_publisher->hasEnded()) {
			// Reading next block
			int32_t blockSize = reader->read();

			// Exiting if no data
			if (blockSize == 0) {
				break;
			}

			// Sending block
			int32_t * memoryBlock = reader->buffer();

			// Put the array into a string
			string data(reinterpret_cast<char *>(memoryBlock), sizeof(int32_t) * blockSize);

			// Send the data
			m_publisher->sendBinary(data);

			//sleep(refreshTimeS());
			usleep(10000);
		}

		cout << "finished reading test file " << testFileName() << endl;
	}
}

std::auto_ptr<boost::thread> RemoteDPPHistogram::startTest(lstdpp::Reader * reader) {

	if (test()) {
		return auto_ptr<thread>(new thread(bind(&RemoteDPPHistogram::testLoop, this, reader)));
	}
	return auto_ptr<thread>();
}

void RemoteDPPHistogram::finishTest(std::auto_ptr<boost::thread> thread) {

	if (test()) {
		// Join the thread
		thread->join();
	}
}

void RemoteDPPHistogram::resetData() {

	// Calculate the total number of channels
	BoardIterator b;

	if (b.end()) {
		totalNumberOfChannels = 0;

	} else {
		int32_t currentNbChannels = (*b).nbChannels;

		++b;
		while (!b.end()) {
			currentNbChannels += (*b).nbChannels;
			++b;
		}

		totalNumberOfChannels = currentNbChannels;
	}

	cout << "totalNumberOfChannels " << totalNumberOfChannels() << endl;

	crateId.resize(totalNumberOfChannels());
	boardId.resize(totalNumberOfChannels());
	channelId.resize(totalNumberOfChannels());
	channelRate.resize(totalNumberOfChannels());
	xData.resize(totalNumberOfChannels());
	histogram.resize(totalNumberOfChannels());
	histogramName.resize(totalNumberOfChannels());
	xLegend.resize(totalNumberOfChannels());
	yLegend.resize(totalNumberOfChannels());

	int32 i = 0;

	// Set the ids
	BoardIterator bc;
	while (!bc.end()) {
		int32 nbChannels = (*bc).nbChannels;
		int32 histogramSize = (1 << BoardDataResolution[(*bc).boardType]);

		for (int32 j = 0; j < nbChannels; ++j) {
			crateId.set(i, bc.crate());
			boardId.set(i, bc.board());
			channelId.set(i, j);
			channelRate.set(i, 0.0);

			delete [] xData(i);
			int32 * values = new int32[histogramSize];
			for (int32 k = 0; k < histogramSize; ++k) {
				values[k] = k;
			}
			xData.set(i, values);
			xData.setSize(i, histogramSize);

			delete [] histogram(i);
			int32 * data = new int32[histogramSize];
			memset(reinterpret_cast<void *>(data), 0, sizeof(int32) * histogramSize);
			histogram.set(i, data);
			histogram.setSize(i, histogramSize);

			string name("ADC");
			name += lexical_cast<string>((int)bc.crate());
			name += ".";
			name += lexical_cast<string>(bc.board());
			name += ".";
			name += lexical_cast<string>(j);
			name += "T";
			histogramName.set(i, name);

			xLegend.set(i, "Energy");
			yLegend.set(i, "Counts");

			++i;
		}

		++bc;
	}
}

}

