/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <cmath>

#include "LEDParamController.h"
#include "controllers/common/family/Families.h"

using namespace std;

namespace stereo_daq {

const string LEDParamController::TYPE = "ledparam_controller";


/*
 * Constructor
 */
LEDParamController::LEDParamController(const string& name) :
	ExperimentController(name),controller::Stoppable(this),Select(this),Set(this) {

	setFamily(family::ACQUISITION,family::SETTING);

	// Init property
	delay.init(this, SAVE, "delay");
	period.init(this, SAVE, "period");
	box1.init(this, NOSAVE, "box1");
	box2.init(this, NOSAVE, "box2");
	box3.init(this, NOSAVE, "box3");
	box4.init(this, NOSAVE, "box4");
	box5.init(this, NOSAVE, "box5");
	pattern.init(this, NOSAVE, "pattern");
	nled.init(this, NOSAVE, "nled");
	dacbox1.init(this, NOSAVE, "dacbox1");
	dacbox2.init(this, NOSAVE, "dacbox2");
	dacbox3.init(this, NOSAVE, "dacbox3");
	dacbox4.init(this, NOSAVE, "dacbox4");
	dacbox5.init(this, NOSAVE, "dacbox5");

	m_ELECDriver.init(this, "driver");

	m_StartActivated = false;

	registerFunction(TYPE);

}

LEDParamController::LEDParamController(const LEDParamController& controller) :
	ExperimentController(controller),controller::Stoppable(this), Select(this),Set(this) {

	delay.copy(this, controller.delay);
	period.copy(this, controller.period);
	box1.copy(this, controller.box1);
	box2.copy(this, controller.box2);
	box3.copy(this, controller.box3);
	box4.copy(this, controller.box4);
	box5.copy(this, controller.box5);
	pattern.copy(this, controller.pattern);
	nled.copy(this, controller.nled);
	dacbox1.copy(this, controller.dacbox1);
	dacbox2.copy(this, controller.dacbox2);
	dacbox3.copy(this, controller.dacbox3);
	dacbox4.copy(this, controller.dacbox4);
	dacbox5.copy(this, controller.dacbox5);

	m_ELECDriver.copy(controller.m_ELECDriver);

	m_StartActivated = false;
	m_ELECDriver->DAQParams.ledBox=0;

}

/*
 * Destructor
 */
LEDParamController::~LEDParamController() {
	// Empty
}

void LEDParamController::postConfiguration() {

	nled=stereo_elec::NBLEDPERBOARD;
	dacbox1.resize(nled());
	dacbox2.resize(nled());
	dacbox3.resize(nled());
	dacbox4.resize(nled());
	dacbox5.resize(nled());

	registerRefresher(delay, &LEDParamController::refreshDelayProperty, this);
	registerRefresher(period, &LEDParamController::refreshPeriodProperty, this);
	registerRefresher(box1, &LEDParamController::refreshBox1Property, this);
	registerRefresher(box2, &LEDParamController::refreshBox2Property, this);
	registerRefresher(box3, &LEDParamController::refreshBox3Property, this);
	registerRefresher(box4, &LEDParamController::refreshBox4Property, this);
	registerRefresher(box5, &LEDParamController::refreshBox5Property, this);
	registerRefresher(pattern, &LEDParamController::refreshPatternProperty, this);
	registerRefresher(dacbox1, &LEDParamController::refreshDACBox1Property, this);
	registerRefresher(dacbox2, &LEDParamController::refreshDACBox2Property, this);
	registerRefresher(dacbox3, &LEDParamController::refreshDACBox3Property, this);
	registerRefresher(dacbox4, &LEDParamController::refreshDACBox4Property, this);
	registerRefresher(dacbox5, &LEDParamController::refreshDACBox5Property, this);

	registerProgression(m_ELECDriver, &LEDParamController::updateProgression, this);


}

void LEDParamController::refreshDelayProperty(int32 value) throw (CannotSetValue) {
	m_ELECDriver->DAQParams.ledPulseDlyVal = (short unsigned) value;
}

void LEDParamController::refreshPeriodProperty(int32 value) throw (CannotSetValue) {
	m_ELECDriver->DAQParams.ledPeriod = (short unsigned) value;
}
void LEDParamController::refreshBox1Property(bool value) throw (CannotSetValue) {
	if(value) m_ELECDriver->DAQParams.ledBox |=  1 ;
	else m_ELECDriver->DAQParams.ledBox &= ~1;
	//cout << dec << "ledbox " << m_ELECDriver->DAQParams.ledBox << endl;
}
void LEDParamController::refreshBox2Property(bool value) throw (CannotSetValue) {
	if(value) m_ELECDriver->DAQParams.ledBox |=  (1 << 1);
	else m_ELECDriver->DAQParams.ledBox &= ~(1<< 1);
	//cout << dec << "ledbox " << m_ELECDriver->DAQParams.ledBox << endl;
}
void LEDParamController::refreshBox3Property(bool value) throw (CannotSetValue) {
	if(value) m_ELECDriver->DAQParams.ledBox |=  (1 << 2);
	else m_ELECDriver->DAQParams.ledBox &= ~(1<< 2);
	//cout << dec << "ledbox " << m_ELECDriver->DAQParams.ledBox << endl;
}
void LEDParamController::refreshBox4Property(bool value) throw (CannotSetValue) {
	if(value) m_ELECDriver->DAQParams.ledBox |=  (1 << 3);
	else m_ELECDriver->DAQParams.ledBox &= ~(1<< 3);
	//cout<< dec  << "ledbox " << m_ELECDriver->DAQParams.ledBox << endl;
}
void LEDParamController::refreshBox5Property(bool value) throw (CannotSetValue) {
	if(value) m_ELECDriver->DAQParams.ledBox |=  (1 << 4);
	else m_ELECDriver->DAQParams.ledBox &= ~(1<< 4);
	//cout << dec << "ledbox " << m_ELECDriver->DAQParams.ledBox << endl;
}
void LEDParamController::refreshPatternProperty(int32 value) throw (CannotSetValue) {
	m_ELECDriver->DAQParams.ledType = (short unsigned) value;
}
void LEDParamController::refreshDACBox1Property(int32 index,int32 value) throw (CannotSetValue) {
	m_ELECDriver->DAQParams.ledDAC.at(0*nled()+index) = (short unsigned) value;
}
void LEDParamController::refreshDACBox2Property(int32 index,int32 value) throw (CannotSetValue) {
	m_ELECDriver->DAQParams.ledDAC[1*nled()+index] = (short unsigned) value;
}
void LEDParamController::refreshDACBox3Property(int32 index,int32 value) throw (CannotSetValue) {
	m_ELECDriver->DAQParams.ledDAC[2*nled()+index] = (short unsigned) value;
}
void LEDParamController::refreshDACBox4Property(int32 index,int32 value) throw (CannotSetValue) {
	m_ELECDriver->DAQParams.ledDAC[3*nled()+index] = (short unsigned) value;
}
void LEDParamController::refreshDACBox5Property(int32 index,int32 value) throw (CannotSetValue) {
	m_ELECDriver->DAQParams.ledDAC[4*nled()+index] = (short unsigned) value;
}

/*
 * updateProgression
 */
void LEDParamController::updateProgression() {
	// Calculate its own progression to manage retry case
	int32 progression = m_ELECDriver->commandProgression();
	if (progression >= 100) {
		progression = 99;
	}
	commandProgression = progression;
}

/*
 * select channels
 */
void LEDParamController::select() {


}

void LEDParamController::set() {

}
void LEDParamController::start() {


}

void LEDParamController::stop() {

}

}

