/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <math.h>

#include "UsbPVCamDriver.h"
#include "UsbPVCamState.h"
#include "RealUsbPVCamDriver.h"
#include "PerfectUsbPVCamDriver.h"
#include "UsbPVCamDef.h"

using namespace std;

namespace usbpvcam {

const std::string UsbPVCamDriver::TYPE = "UsbPVCam";

/*
 * Constructor
 */
UsbPVCamDriver::UsbPVCamDriver(const string& name) :
		acquisition::NoTofAcquisitionCommon(name), acquisition::TimeGateCommon(name), acquisition::DetectorCommon(name) {

	registerStates(new RealUsbPVCamDriver(this), new PerfectUsbPVCamDriver(this), new PerfectUsbPVCamDriver(this));

	/*
	 * Init the device command list
	 */
	initCommand(driver::INIT_COMMAND);
	initCommand(driver::READ_COMMAND);
	initCommand(driver::SYNCHRONIZE_READ_COMMAND);
	initCommand(driver::START_COMMAND);
	initCommand(driver::RESUME_COMMAND);
	initCommand(driver::PAUSE_COMMAND);
	initCommand(driver::STOP_COMMAND);
	initCommand(driver::STATUS_COMMAND);
	initCommand(driver::READ_INFOS_COMMAND);
	initCommand(DetectorCommon::CLEAR_COMMAND);
	initCommand(DetectorCommon::REGROUP_COMMAND);
	initCommand(AcquisitionCommon::WRITE_PARAMETERS_COMMAND);

	// Init functions
	registerParentFunction(NONE_FUNCTION);
	registerFunction(NONE_FUNCTION);

	deviceType = ROOT_DEVICE_TYPE_DEVICE_CONTAINER;
	m_SynchroniseReadActivated = false;
	m_SynchroniseReadTerminated = false;
	m_DataShort = NULL;

	registerObserverCommand(COMMAND_STATUS_DEVICE_CONTAINER, 50);
}

/*
 * Destructor
 */
UsbPVCamDriver::~UsbPVCamDriver() {
}

/*
 * refreshDataSizeProperty
 */
void UsbPVCamDriver::refreshDataSizeProperty(int32 value) throw (CannotSetValue) {
	if (m_DataShort != NULL) {
		delete[] m_DataShort;
	}
	m_DataShort = new uint16[value];
	if (m_DataShort == NULL) {
		sendErrorEvent(NOTENOUGH_MEMORY_ERROR, "Can't create data table");
		throw CannotSetValue();
	}
	DetectorCommon::refreshDataSizeProperty(value);
}

/*
 * refreshModeProperty
 */
void UsbPVCamDriver::refreshModeProperty(int32 value) throw (CannotSetValue) {
	if (value != acquisition::GateCommon::TIME_MODE) {
		throw CannotSetValue();
	}
}

/*
 * execute
 */
void UsbPVCamDriver::execute(const std::string& aCommand) {

	if ((aCommand != driver::PAUSE_COMMAND) && (aCommand != driver::STOP_COMMAND) && (aCommand != driver::STATUS_COMMAND)
			&& (aCommand != driver::READ_COMMAND)) {
		commandProgression = PROGRESSION_UNKNOWNSTATE_DEVICE_CONTAINER;
	}
	UsbPVCamState* currentState = dynamic_cast<UsbPVCamState*>(getCurrentState());
	// Check command
	if (aCommand == driver::INIT_COMMAND) {
		// Init command
		currentState->init();
	} else if (aCommand == driver::READ_COMMAND) {
		// Read command
		currentState->read();
	} else if (aCommand == driver::SYNCHRONIZE_READ_COMMAND) {
		// Read command
		m_SynchroniseReadActivated = true;
		m_SynchroniseReadTerminated = false;
		currentState->synchroniseRead();
	} else if (aCommand == driver::START_COMMAND) {
		// Start command
		startActivated = true;
		currentState->start();
	} else if (aCommand == driver::RESUME_COMMAND) {
		// Continue command
		currentState->resume();
	} else if (aCommand == driver::PAUSE_COMMAND) {
		// Stop command
		currentState->pause();
	} else if (aCommand == driver::STOP_COMMAND) {
		// Stop command
		currentState->stop();
		startActivated = false;
	} else if (aCommand == driver::STATUS_COMMAND) {
		// Status command
		currentState->readStatus();
	} else if (aCommand == driver::READ_INFOS_COMMAND) {
		// Info command
		currentState->readInfos();
	} else if (aCommand == DetectorCommon::CLEAR_COMMAND) {
		// Move command
		sum = 0;
		currentState->clear();
	} else if (aCommand == AcquisitionCommon::WRITE_PARAMETERS_COMMAND) {
		// Write command
		currentState->writeParam();
	} else {
		//Error bad command
		//Todo , normaly it will be detect it device action driver
	}
}

/*
 * calculateProgression
 */
void UsbPVCamDriver::calculateProgression() {
	int32 progression = 0;
	float64 setpoint = 0.0;
	float64 actual = 0.0;

	setpoint = time.setpoint();
	actual = time();

	if (setpoint == 0.0) {
		commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
	} else {
		progression = (int32) (actual / setpoint * PROGRESSION_RATIO_DEVICE_CONTAINER);
		// Be sure that progression > 0%
		if (progression < PROGRESSION_UNKNOWNSTATE_DEVICE_CONTAINER)
			progression = 0;
		// Be sure that only Phased status put 100%
		if (progression >= PROGRESSION_END_DEVICE_CONTAINER)
			progression = PROGRESSION_END_DEVICE_CONTAINER - 1;
		commandProgression = progression;
	}
}

/*
 * UsbPVCamDriver
 */
void UsbPVCamDriver::computeGateStatus(int32 value) {
	int32 valueout = 0;
	if ((value == pvcam::EXPOSURE_IN_PROGRESS) || (value == pvcam::READOUT_IN_PROGRESS)
			|| (value == pvcam::ACQUISITION_IN_PROGRESS)) {
		valueout |= GateCommon::RUNNING_STATUS;
	} else if (value == pvcam::READOUT_COMPLETE) {
		valueout |= GateCommon::END_STATUS;
	}
	gateStatus.update(valueout);
}

/*
 * UsbPVCamDriver
 */
void UsbPVCamDriver::computeDetectorStatus(int32 value) {
	int32 valueout = 0;
	detectorStatus.update(valueout);

}

/*
 * UsbPVCamDriver
 */
void UsbPVCamDriver::computeAcquisitionStatus(int32 value) {
	int32 valueout = 0;
	acquisitionStatus.update(valueout);
}

}

